import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import QuantileTransformer


"""Grouped Quantile Featurizer"""

__author__ = "david26694"


class GroupedQuantileTransformer(BaseEstimator, TransformerMixin):
    """
    Computes the group quantile of a numeric feature with respect to a categorical feature.

    For instance, if each datum is an apartment, and we have both the price and the city,
    this feature tries to model how expensive is an apartment in its city. The most
    expensive apartment in the city will score 1, and the cheapest will score 0.

    It is equivalent at what it is done in:
    https://stackoverflow.com/questions/33899369/ranking-order-per-group-in-pandas

    Parameters
    ----------

    feature_mapping: dict
        mapping from numeric variables to categories that want to be used as groups.
    n_quantiles: int
        number of quantiles per category.
    subsample: int
        Maximum number of samples used to estimate the quantiles for computational efficiency.
    random_state: Any
        Determines random number generation for subsampling and smoothing noise. Please see ``subsample`` for more details.
        Pass an int for reproducible results across multiple function calls. See :term:`Glossary  `
    copy: bool
        Set to False to perform inplace transformation and avoid a copy (if the input is already a numpy array).

    Example
    -------
    >>> from sktools import GroupedQuantileTransformer
    >>> import pandas as pd
    >>> X = pd.DataFrame(
    >>>         {
    >>>             "price": [1, 2, 3, 3, 2, 10, 0],
    >>>             "city": ["a", "a", "a", "b", "b", None, None],
    >>>         }
    >>>     )
    >>> featurizer = GroupedQuantileTransformer(feature_mapping={"price": "city"})
    >>> print(featurizer.fit_transform(X).columns)
    Index(['price', 'city', 'price_quantile_city'], dtype='object')

    """

    def __init__(
        self,
        feature_mapping,
        handle_missing="value",
        n_quantiles=1000,
        subsample=int(1e5),
        random_state=None,
        copy=True,
    ):
        self.transformer_dict = {}
        self.feature_mapping = feature_mapping
        self.transformers = None
        self.n_quantiles = n_quantiles
        self.subsample = subsample
        self.random_state = random_state
        self.copy = copy
        self.handle_missing = handle_missing

    def fit(self, X, y=None):

        for col, group in self.feature_mapping.items():

            self.transformer_dict[group] = {}
            categories = X[group].unique()

            for category in categories:
                # Regular case - non-nulls -> create a quantile transformer
                # and fit it with data in that category
                if category is not None:
                    self.transformer_dict[group][
                        category
                    ] = QuantileTransformer(
                        n_quantiles=self.n_quantiles,
                        subsample=self.subsample,
                        random_state=self.random_state,
                        copy=self.copy,
                    )

                    x_category = X[X[group] == category]
                    self.transformer_dict[group][category].fit(
                        x_category.loc[:, [col]]
                    )

            # Non-regular case -> impute missings by taking the whole distribution
            if self.handle_missing == "value":
                self.transformer_dict[group][np.nan] = QuantileTransformer(
                    n_quantiles=self.n_quantiles,
                    subsample=self.subsample,
                    random_state=self.random_state,
                    copy=self.copy,
                )

                x_category = X
                self.transformer_dict[group][np.nan].fit(
                    x_category.loc[:, [col]]
                )

        return self

    def transform(self, X):

        if self.copy:
            X = X.copy()

        for col, group in self.feature_mapping.items():

            transform_feature_name = f"{col}_quantile_{group}"
            X[transform_feature_name] = np.zeros(X.shape[0])

            categories = X[group].unique()
            fit_categories = self.transformer_dict[group].keys()

            for category in categories:

                # Easy case - regular category
                # Just use transformer_dict to estimate to establish quantile
                # for the features
                if category in fit_categories and category is not np.nan:
                    x_category = X[X[group] == category]
                    x_col = x_category.loc[:, [col]]
                    transformer = self.transformer_dict[group][category]
                    x_transform = transformer.transform(x_col)

                    X.loc[
                        X[group] == category, transform_feature_name
                    ] = x_transform

                # New categories or nulls -> use default transformer
                else:
                    # Keep new and null
                    nonnull_fit_cats = set(fit_categories).difference([np.nan])
                    other_cats_condition = ~X[group].isin(nonnull_fit_cats)
                    x_category = X[other_cats_condition]

                    # Use default transformer
                    transformer = self.transformer_dict[group][np.nan]
                    x_transform = transformer.transform(
                        x_category.loc[:, [col]]
                    )

                    # Assign to X
                    X.loc[
                        other_cats_condition, transform_feature_name
                    ] = x_transform

        return X


class PercentileGroupFeaturizer(BaseEstimator, TransformerMixin):
    """
        Creates features establishing a relationship between a numeric and a categorical feature,
        by using a given percentile of the numeric feature in each cateogry.

        For instance, if each datum is an apartment, and we have both the price and the city,
        if we use the percentile 50 the features model how expensive is an apartment
        with respect to the median in the city.

    Parameters
    ----------

    feature_mapping: dict
        mapping from numeric variables to categories that want to be used as groups.
    percentile: int
        percentile used to compute features
    create_features: bool
        If false, it just computes percentiles by category
    handle_missing: str
        options are 'return_nan' and 'value', defaults to 'value', which uses the global quantile.
    handle_unknown: str
        options are 'return_nan' and 'value', defaults to 'value', which uses the global quantile.

    Example
    -------
    >>> from sktools import PercentileGroupFeaturizer
    >>> import pandas as pd
    >>> X = pd.DataFrame(
    >>>         {
    >>>             "price": [1, 2, 3, 3, 2, 10, 0],
    >>>             "city": ["a", "a", "a", "b", "b", None, None],
    >>>         }
    >>>     )
    >>> featurizer = PercentileGroupFeaturizer(
    >>>     feature_mapping={"price": "city"}
    >>> )
    >>> print(featurizer.fit_transform(X).columns)
    Index(['price', 'city', 'p50_price_city', 'diff_p50_price_city',
           'relu_diff_p50_price_city', 'ratio_p50_price_city'],
          dtype='object')


    """

    def __init__(
        self,
        feature_mapping,
        percentile=50,
        create_features=True,
        handle_missing="value",
        handle_unknown="value",
    ):
        self.feature_mapping = feature_mapping
        self.percentile = percentile
        self.create_features = create_features
        self.handle_missing = handle_missing
        self.handle_unknown = handle_unknown
        self.saved_percentiles = {}

    def fit(self, X, y=None):

        for col, group in self.feature_mapping.items():

            pctl_col_name = f"p{self.percentile}_{col}_{group}"

            # Create percentile by group
            pctl_df = (
                X.groupby(group, as_index=False)
                .agg({col: lambda x: x.quantile(self.percentile / 100)})
                .rename(columns={col: pctl_col_name})
            )

            # Regular handle missing -> add global percentile to missing
            if self.handle_missing == "value":
                global_pctl = X[col].agg(
                    lambda x: x.quantile(self.percentile / 100)
                )

                global_pctl_df = pd.DataFrame(
                    {group: np.nan, pctl_col_name: [global_pctl]}
                )

                pctl_df = pd.concat([pctl_df, global_pctl_df])

            self.saved_percentiles[col] = pctl_df

        return self

    def transform(self, X):

        X = X.copy()

        for col, group in self.feature_mapping.items():

            X = X.merge(self.saved_percentiles[col], on=group, how="left")
            pctl_col_name = f"p{self.percentile}_{col}_{group}"

            # First assign percentiles to non-trivial cases, which are
            # new categories
            if (
                self.handle_unknown == "value"
                and self.handle_missing == "value"
            ):
                groups_fit = self.saved_percentiles[col][group]
                new_condition = (~X[group].isin(groups_fit)) & X[
                    pctl_col_name
                ].isnull()

                x_fit = self.saved_percentiles[col]
                imputation = float(
                    x_fit.loc[x_fit[group].isnull()][pctl_col_name]
                )
                X.loc[new_condition, pctl_col_name] = imputation

            # Then trivially create features
            if self.create_features:
                diff_name = f"diff_p{self.percentile}_{col}_{group}"
                relu_diff_name = f"relu_diff_p{self.percentile}_{col}_{group}"
                ratio_name = f"ratio_p{self.percentile}_{col}_{group}"

                X[diff_name] = X[col] - X[pctl_col_name]
                X[relu_diff_name] = X[diff_name].clip(0, np.inf)
                X[ratio_name] = X[col] / X[pctl_col_name]

        return X


class MeanGroupFeaturizer(BaseEstimator, TransformerMixin):
    """
        Creates features establishing a relationship between a numeric and a categorical feature,
        by using the mean of the numeric feature in each cateogry.

        For instance, if each datum is an apartment, and we have both the price and the city,
        the features model how expensive is an apartment with respect to the mean in the city.

    Parameters
    ----------

    feature_mapping: dict
        mapping from numeric variables to categories that want to be used as groups.
    percentile: int
        percentile used to compute features
    create_features: bool
        If false, it just computes percentiles by category
    handle_missing: str
        options are 'return_nan' and 'value', defaults to 'value', which uses the global quantile.
    handle_unknown: str
        options are 'return_nan' and 'value', defaults to 'value', which uses the global quantile.

    Example
    -------
    >>> from sktools import MeanGroupFeaturizer
    >>> import pandas as pd
    >>> X = pd.DataFrame(
    >>>         {
    >>>             "price": [1, 2, 3, 3, 2, 10, 0],
    >>>             "city": ["a", "a", "a", "b", "b", None, None],
    >>>         }
    >>>     )
    >>> featurizer = MeanGroupFeaturizer(
    >>>     feature_mapping={"price": "city"}
    >>> )
    >>> print(featurizer.fit_transform(X).columns)
    Index(['price', 'city', 'mean_price_city', 'diff_mean_price_city',
           'relu_diff_mean_price_city', 'ratio_mean_price_city'],
          dtype='object')


    """

    def __init__(
        self,
        feature_mapping,
        create_features=True,
        handle_missing="value",
        handle_unknown="value",
    ):
        self.feature_mapping = feature_mapping
        self.create_features = create_features
        self.handle_missing = handle_missing
        self.handle_unknown = handle_unknown
        self.saved_mean = {}

    def fit(self, X, y=None):

        for col, group in self.feature_mapping.items():

            mean_col_name = f"mean_{col}_{group}"

            # Create percentile by group
            mean_df = (
                X.groupby(group, as_index=False)
                .agg({col: lambda x: x.mean()})
                .rename(columns={col: mean_col_name})
            )

            # Regular handle missing -> add global percentile to missing
            if self.handle_missing == "value":
                global_mean = X[col].agg(lambda x: x.mean())

                global_mean_df = pd.DataFrame(
                    {group: np.nan, mean_col_name: [global_mean]}
                )

                mean_df = pd.concat([mean_df, global_mean_df])

            self.saved_mean[col] = mean_df

        return self

    def transform(self, X):

        X = X.copy()

        for col, group in self.feature_mapping.items():

            X = X.merge(self.saved_mean[col], on=group, how="left")
            mean_col_name = f"mean_{col}_{group}"

            # First assign percentiles to non-trivial cases, which are
            # new categories
            if (
                self.handle_unknown == "value"
                and self.handle_missing == "value"
            ):
                groups_fit = self.saved_mean[col][group]
                new_condition = (~X[group].isin(groups_fit)) & X[
                    mean_col_name
                ].isnull()

                x_fit = self.saved_mean[col]
                imputation = float(
                    x_fit.loc[x_fit[group].isnull()][mean_col_name]
                )
                X.loc[new_condition, mean_col_name] = imputation

            # Then trivially create features
            if self.create_features:
                diff_name = f"diff_mean_{col}_{group}"
                relu_diff_name = f"relu_diff_mean_{col}_{group}"
                ratio_name = f"ratio_mean_{col}_{group}"

                X[diff_name] = X[col] - X[mean_col_name]
                X[relu_diff_name] = X[diff_name].clip(0, np.inf)
                X[ratio_name] = X[col] / X[mean_col_name]

        return X



"""Main module."""

from sklearn.base import BaseEstimator, TransformerMixin


class MatrixDenser(BaseEstimator, TransformerMixin):
    """
    Converts matrix to dense.

    Useful when doing an union between dense and sparse matrices.

    Example
    -------
    >>> from sktools import MatrixDenser
    >>> from scipy.sparse import csr_matrix
    >>> import numpy as np
    >>> sparse_matrix = csr_matrix((3, 4), dtype=np.int8)
    >>> dense_matrix = MatrixDenser().fit_transform(sparse_matrix)
    >>> print(dense_matrix)
    [[0 0 0 0]
     [0 0 0 0]
     [0 0 0 0]]
    """

    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.toarray()



#!/usr/bin/env python

"""Tests for `sktools` package."""

import unittest

import sktools
import pandas as pd
from scipy.sparse import csr_matrix
from category_encoders import MEstimateEncoder
import numpy as np

from rich.console import Console
from rich.traceback import install

console = Console()
install()


class TestTypeSelector(unittest.TestCase):
    """Tests for type selector."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "char_col": ["a", "b", "c"],
                "other_char_col": ["d", "e", "f"],
            }
        )

    def run_test_type(self, dtype):
        """
        This test applies the fit and transform methods with a given type.
        It then asserts that the type of each column is the same as the input


        @param dtype: Transformer is built using this type
        """

        type_transformer = sktools.TypeSelector(dtype)
        type_cols = type_transformer.fit_transform(self.df)
        output_types = type_cols.dtypes

        for type_col in output_types:
            self.assertEqual(type_col, dtype)

    def test_integer_works(self):
        self.run_test_type("int64")

    def test_object_works(self):
        self.run_test_type("object")

    def test_float_works(self):
        self.run_test_type("float64")


class TestItemSelector(unittest.TestCase):
    """Tests for item selector."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "char_col": ["a", "b", "c"],
                "other_char_col": ["d", "e", "f"],
            }
        )

    def test_select_items(self):
        """
        Check that item selector works for each column, that is:
        * Name is the same as the column of the dataframe
        * Values are the same as the values in the dataframe
        """

        for col in self.df.columns:
            col_transformer = sktools.ItemSelector(col)
            output_transformer = col_transformer.fit_transform(self.df)
            output_column = output_transformer.name

            self.assertEqual(output_column, col)

            self.assertTrue(
                (output_transformer == self.df[col]).all(),
                "Not all values of the series are equal",
            )


class TestMatrixDenser(unittest.TestCase):
    """Tests for item selector."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.sparse_matrix = csr_matrix((3, 4), dtype=np.int8)

    def test_zero_matrix(self):
        dense_matrix = sktools.MatrixDenser().fit_transform(self.sparse_matrix)

        expected_dense = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], dtype=np.int8
        )

        self.assertTrue(
            (dense_matrix == expected_dense).all(), "Not all values are 0"
        )


class TestEmptyExtractor(unittest.TestCase):
    """Tests for item selector."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.df = pd.DataFrame(
            {
                "int_col": [1, 2, np.NaN],
                "float_col": [1.1, np.NaN, 3.3],
                "char_col": [np.NaN, "b", "c"],
                "other_char_col": ["d", "e", "f"],
            }
        )
        self.expected_output = pd.DataFrame(
            {
                "int_col": [1, 2, np.NaN],
                "float_col": [1.1, np.NaN, 3.3],
                "char_col": [np.NaN, "b", "c"],
                "other_char_col": ["d", "e", "f"],
                "int_col_na": [False, False, True],
                "float_col_na": [False, True, False],
                "char_col_na": [True, False, False],
                "other_char_col_na": [False, False, False],
            }
        )

    def test_defaults(self):
        pd.testing.assert_frame_equal(
            sktools.IsEmptyExtractor().fit_transform(self.df),
            self.expected_output.drop("other_char_col_na", axis=1),
        )

    def test_non_delete(self):
        pd.testing.assert_frame_equal(
            sktools.IsEmptyExtractor(keep_trivial=True).fit_transform(self.df),
            self.expected_output,
        )


class TestPercentileEncoder(unittest.TestCase):
    """Tests for percentile encoder."""

    def setUp(self):
        """Create dataframe with categories and a target variable"""

        self.df = pd.DataFrame(
            {"categories": ["a", "b", "c", "a", "b", "c", "a", "b"]}
        )
        self.target = np.array([1, 2, 0, 4, 5, 0, 6, 7])

    def test_median_works(self):
        """
        Expected output of percentile 50 in df:
            - a median is 4 (a values are 1, 4, 6)
            - b median is 5 (b values are 2, 5, 7)
            - c median is 0 (c values are 0)
        """

        expected_output_median = pd.DataFrame(
            {"categories": [4.0, 5, 0, 4, 5, 0, 4, 5]}
        )

        pd.testing.assert_frame_equal(
            sktools.PercentileEncoder(percentile=50).fit_transform(
                self.df, self.target
            ),
            expected_output_median,
        )

    def test_max_works(self):
        """
        Expected output of percentile 100 in df:
            - a max is 6
            - b max is 7
            - c max is 0
        """
        expected_output_max = pd.DataFrame(
            {"categories": [6.0, 7, 0, 6, 7, 0, 6, 7]}
        )

        pd.testing.assert_frame_equal(
            sktools.PercentileEncoder(percentile=100).fit_transform(
                self.df, self.target
            ),
            expected_output_max,
        )

    def test_new_category(self):
        """
        The global median of the target is 3. If new categories are passed to
        the transformer, then the output should be 3
        """
        transformer_median = sktools.PercentileEncoder(percentile=50)
        transformer_median.fit(self.df, self.target)

        new_df = pd.DataFrame({"categories": ["d", "e"]})

        new_medians = pd.DataFrame({"categories": [3.0, 3.0]})

        pd.testing.assert_frame_equal(
            transformer_median.transform(new_df), new_medians
        )


class TestNestedTargetEncoder(unittest.TestCase):
    """Tests for nested target encoder."""

    def setUp(self):
        """Create dataframe with categories and a target variable"""

        self.col = "col_1"
        self.parent_col = "parent_col_1"
        self.X = pd.DataFrame(
            {
                self.col: ["a", "a", "b", "b", "b", "c", "c", "d", "d", "d"],
                self.parent_col: [
                    "e",
                    "e",
                    "e",
                    "e",
                    "e",
                    "f",
                    "f",
                    "f",
                    "f",
                    "f",
                ],
            }
        )

        self.X_array = pd.DataFrame(
            {
                self.col: ["a", "a", "b", "b", "b", "c", "c", "d", "d", "d"],
                self.parent_col: [
                    "e",
                    "e",
                    "e",
                    "e",
                    "e",
                    "f",
                    "f",
                    "f",
                    "f",
                    "f",
                ],
            }
        ).values

        self.y = pd.Series([1, 2, 3, 1, 2, 4, 4, 5, 4, 4.5])

        self.parent_means = list(self.y.groupby(self.X[self.parent_col]).mean())
        self.parents = ["e", "f"]

    def test_parent_prior(self):
        """
        Simple case:
        There is no prior from the global to the group mean (m_prior = 0).
        As the m_parent is 1, the mean for group a is (as mean_group_e = 1.8):
        (1 + 2 + mean_group_e ) / 3 = (1 + 2 + 1.8) / 3 = 1.6
        The same works for b, c and d
        """
        expected_output = pd.DataFrame(
            dict(
                col_1=[1.6, 1.6, 1.95, 1.95, 1.95, 4.1, 4.1, 4.45, 4.45, 4.45],
                parent_col_1=self.X[self.parent_col],
            )
        )

        te = sktools.NestedTargetEncoder(
            cols=self.col,
            feature_mapping=dict(col_1=self.parent_col),
            m_prior=0,
        )
        pd.testing.assert_frame_equal(
            te.fit_transform(self.X, self.y), expected_output
        )

    def test_numpy_array(self):
        """
        Check that nested target encoder also works for numpy arrays
        """
        expected_output = pd.DataFrame(
            dict(
                col_1=[1.6, 1.6, 1.95, 1.95, 1.95, 4.1, 4.1, 4.45, 4.45, 4.45],
                parent_col_1=self.X[self.parent_col],
            )
        ).values

        te = sktools.NestedTargetEncoder(
            cols=0, feature_mapping={0: 1}, m_prior=0
        )

        te.fit(self.X_array, self.y)
        output = te.transform(self.X_array).values

        np.testing.assert_almost_equal(output[:, 0], expected_output[:, 0])

        np.testing.assert_equal(output[:, 1], expected_output[:, 1])

    def test_no_parent(self):
        """
        When using no priors, the functionalities should be the same as for
        m estimator.
        """

        te = sktools.NestedTargetEncoder(
            cols=self.col,
            feature_mapping=dict(col_1=self.parent_col),
            m_prior=0,
            m_parent=0,
        )

        m_te = MEstimateEncoder(cols=self.col, m=0)
        pd.testing.assert_frame_equal(
            te.fit_transform(self.X, self.y), m_te.fit_transform(self.X, self.y)
        )

    def test_unknown_missing_imputation(self):
        """
        When new categories and unknown values are given, we expect the encoder
        to give the parent means (at least with default configuration).
        """

        # First two rows are new categories
        # Last two rows are missing values
        # Parents are e, f, e, f
        new_x = pd.DataFrame(
            {
                self.col: ["x", "y", np.NaN, np.NaN],
                self.parent_col: self.parents + self.parents,
            }
        )

        # We expect to get parent means
        expected_output_df = pd.DataFrame(
            {
                self.col: self.parent_means + self.parent_means,
                self.parent_col: self.parents + self.parents,
            }
        )

        te = sktools.NestedTargetEncoder(
            cols=self.col,
            feature_mapping=dict(col_1=self.parent_col),
            m_prior=0,
        )

        te.fit(self.X, self.y)

        pd.testing.assert_frame_equal(te.transform(new_x), expected_output_df)

    def test_missing_na(self):
        """
        When new categories and unknown values are given, we expect the encoder
        to give the parent means. If we specify return_nan, we want it to
        return nan
        """

        # First two rows are new categories
        # Last two rows are missing values
        # Parents are e, f, e, f
        new_x = pd.DataFrame(
            {
                self.col: ["x", "y", np.nan, np.nan],
                self.parent_col: self.parents + self.parents,
            }
        )

        # In the transformer we specify unknown -> return nan
        # We expect to get:
        # - nan for the unknown
        # - parent means for the missing
        expected_output_df = pd.DataFrame(
            {
                self.col: [np.nan, np.nan] + self.parent_means,
                self.parent_col: self.parents + self.parents,
            }
        )

        te = sktools.NestedTargetEncoder(
            cols=self.col,
            feature_mapping=dict(col_1=self.parent_col),
            m_prior=0,
            handle_missing="value",
            handle_unknown="return_nan",
        )

        te.fit(self.X, self.y)

        pd.testing.assert_frame_equal(te.transform(new_x), expected_output_df)

    def test_all_missing(self):
        """
        If everything's missing or unknow , we expect by default to return
        global mean
        """
        new_x = pd.DataFrame(
            {
                self.col: ["x", np.nan, "x", np.nan],
                self.parent_col: ["z", "z", np.nan, np.nan],
            }
        )

        te = sktools.NestedTargetEncoder(
            cols=self.col,
            feature_mapping=dict(col_1=self.parent_col),
            m_prior=0,
        )

        te.fit(self.X, self.y)

        expected_output_df = pd.DataFrame(
            {
                self.col: self.y.mean(),
                self.parent_col: ["z", "z", np.nan, np.nan],
            }
        )

        pd.testing.assert_frame_equal(te.transform(new_x), expected_output_df)


class TestGroupQuantile(unittest.TestCase):
    """Tests for group quantile."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.X = pd.DataFrame(
            {
                "x": [1, 2, 3, 2, 20, 0, 10],
                "group": ["a", "a", "b", "b", None, None, "c"],
            }
        )

        self.new_X = pd.DataFrame({"x": [100.0], "group": ["d"]})

        # TODO: only 1 class should give .5 -> smoothing?
        self.output = self.X.copy().assign(
            x_quantile_group=[0.0, 1, 1, 0, 1, 0, 0.5]
        )
        self.new_output = self.new_X.copy().assign(x_quantile_group=[1.0])

    def test_basic_example(self):
        groupedquantile = sktools.GroupedQuantileTransformer(
            feature_mapping={"x": "group"}
        )
        groupedquantile.fit(self.X)

        transformation = groupedquantile.transform(self.X)
        pd.testing.assert_frame_equal(transformation, self.output)

    def test_unknown(self):
        groupedquantile = sktools.GroupedQuantileTransformer(
            feature_mapping={"x": "group"}
        )
        groupedquantile.fit(self.X)

        transformation = groupedquantile.transform(self.new_X)
        pd.testing.assert_frame_equal(transformation, self.new_output)


class TestGroupQuantileFeaturizer(unittest.TestCase):
    """Tests for group quantile."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.X = pd.DataFrame(
            {
                "x": [1, 2, 3, 2, 20, 0, 10],
                "group": ["a", "a", "b", "b", None, None, "c"],
            }
        )

        self.new_X = pd.DataFrame(
            {"x": [1, 2, 3, 4, 5], "group": ["a", "b", "c", "d", None]}
        )

        self.output = self.X.copy().assign(
            p50_x_group=[1.5, 1.5, 2.5, 2.5, 2.0, 2.0, 10],
            diff_p50_x_group=[-0.5, 0.5, 0.5, -0.5, 18.0, -2, 0],
            relu_diff_p50_x_group=[0, 0.5, 0.5, 0, 18.0, 0.0, 0],
            ratio_p50_x_group=[2.0 / 3, 4.0 / 3, 1.2, 0.8, 10, 0.0, 1.0],
        )

        self.missing_output = self.X.copy().assign(
            p50_x_group=[1.5, 1.5, 2.5, 2.5, None, None, 10],
            diff_p50_x_group=[-0.5, 0.5, 0.5, -0.5, None, None, 0],
        )

        self.new_output = self.new_X.copy().assign(
            p50_x_group=[1.5, 2.5, 10, 2.0, 2.0]
        )

    def test_basic_featurizer(self):
        featurizer = sktools.PercentileGroupFeaturizer(
            feature_mapping={"x": "group"}
        )

        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(featurizer.transform(self.X), self.output)

    def test_missing(self):
        featurizer = sktools.PercentileGroupFeaturizer(
            feature_mapping={"x": "group"}, handle_missing="return_nan"
        )
        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(
            featurizer.transform(self.X).iloc[:, :4], self.missing_output
        )

    def test_new_input(self):
        featurizer = sktools.PercentileGroupFeaturizer(
            feature_mapping={"x": "group"}
        )
        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(
            featurizer.transform(self.new_X).iloc[:, 0:3], self.new_output
        )


class TestMeanFeaturizer(unittest.TestCase):
    """Tests for mean quantile."""

    def setUp(self):
        """Create dataframe with different column types"""
        self.X = pd.DataFrame(
            {
                "x": [1, 2, 3, 2, 10, 0, 10],
                "group": ["a", "a", "b", "b", None, None, "c"],
            }
        )

        self.new_X = pd.DataFrame(
            {"x": [1, 2, 3, 4, 5], "group": ["a", "b", "c", "d", None]}
        )

        self.output = self.X.copy().assign(
            mean_x_group=[1.5, 1.5, 2.5, 2.5, 4.0, 4.0, 10],
            diff_mean_x_group=[-0.5, 0.5, 0.5, -0.5, 6, -4, 0],
            relu_diff_mean_x_group=[0, 0.5, 0.5, 0, 6, 0.0, 0],
            ratio_mean_x_group=[2.0 / 3, 4.0 / 3, 1.2, 0.8, 10.0 / 4, 0.0, 1.0],
        )

        self.new_output = self.new_X.copy().assign(
            mean_x_group=[1.5, 2.5, 10, 4.0, 4.0]
        )

        self.missing_output = self.X.copy().assign(
            mean_x_group=[1.5, 1.5, 2.5, 2.5, None, None, 10],
            diff_mean_x_group=[-0.5, 0.5, 0.5, -0.5, None, None, 0],
        )

    def test_basic_featurizer(self):
        featurizer = sktools.MeanGroupFeaturizer(feature_mapping={"x": "group"})

        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(featurizer.transform(self.X), self.output)

    def test_missing(self):
        featurizer = sktools.MeanGroupFeaturizer(
            feature_mapping={"x": "group"}, handle_missing="return_nan"
        )
        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(
            featurizer.transform(self.X).iloc[:, :4], self.missing_output
        )

    def test_new_input(self):
        featurizer = sktools.MeanGroupFeaturizer(feature_mapping={"x": "group"})
        featurizer.fit(self.X)

        pd.testing.assert_frame_equal(
            featurizer.transform(self.new_X).iloc[:, 0:3], self.new_output
        )



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class TypeSelector(BaseEstimator, TransformerMixin):
    """
    Transformer that filters a type of columns of a given data frame. This can
    be useful if we want to treat numeric and object columns differently


    Parameters
    ----------
    dtype : required
        The type we want to filter

    Example
    -------
    >>> from sktools import TypeSelector
    >>> import pandas as pd
    >>> X = pd.DataFrame(
    >>>         {
    >>>             "price": [1., 2., 3.],
    >>>             "city": ["a", "a", "b"]
    >>>         }
    >>>     )
    >>> selector = TypeSelector(
    >>>     dtype='float'
    >>> )
    >>> print(selector.fit_transform(X))
        price
    0    1.0
    1    2.0
    2    3.0
    """

    def __init__(self, dtype):
        self.dtype = dtype

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        assert isinstance(X, pd.DataFrame)
        return X.select_dtypes(include=[self.dtype])


class ItemSelector(BaseEstimator, TransformerMixin):
    """For data grouped by feature, select subset of data at a provided key.

    The data is expected to be stored in a 2D data structure, where the first
    index is over features and the second is over samples.  i.e.

    >>> len(data[key]) == n_samples

    Please note that this is the opposite convention to scikit-learn feature
    matrices (where the first index corresponds to sample).

    ItemSelector only requires that the collection implement getitem
    (data[key]).  Examples include: a dict of lists, 2D numpy array, Pandas
    DataFrame, numpy record array, etc.

    >>> data = {'a': [1, 5, 2, 5, 2, 8],
                'b': [9, 4, 1, 4, 1, 3]}
    >>> ds = ItemSelector(key='a')
    >>> data['a'] == ds.transform(data)

    ItemSelector is not designed to handle data grouped by sample.  (e.g. a
    list of dicts).  If your data is structured this way, consider a
    transformer along the lines of `sklearn.feature_extraction.DictVectorizer`.

    Parameters
    ----------
    key : hashable, required
        The key corresponding to the desired value in a mappable.
    """

    def __init__(self, key):
        self.key = key

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X[self.key]



import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


class IsEmptyExtractor(BaseEstimator, TransformerMixin):
    """
    Transformer that adds columns indicating wether columns have NaN values
    in a row

    Parameters
    ----------
    keep_trivial:
        If a column doesn't have NaN, don't add the column
    cols:
        List of columns to transform. If None, all columns are transformed.
        It only works if input is a DataFrame
    Example
    -------
    >>> from sktools import IsEmptyExtractor
    >>> import pandas as pd
    >>> import numpy as np
    >>> X = pd.DataFrame(
    >>>     {
    >>>         "x": ["a", "b", np.nan],
    >>>         "y": ["c", np.nan, "d"]
    >>>     }
    >>> )
    >>> IsEmptyExtractor().fit_transform(X)
         x    y   x_na   y_na
    0    a    c  False  False
    1    b  NaN  False   True
    2  NaN    d   True  False


    """

    def __init__(self, keep_trivial=False, cols=None):

        self.keep_trivial = keep_trivial
        self.cols = cols

    def fit(self, X, y=None):
        return self

    def transform_data_frame(self, X):
        """
        Transform method in case of receiving a pandas data frame
        """

        new_x = X.copy()

        if self.cols is None:
            selected_columns = X.columns
        else:
            selected_columns = self.cols

        for column in selected_columns:

            new_column_name = f"{column}_na"

            is_na = X[column].isna()

            if any(is_na) or self.keep_trivial:
                new_x[new_column_name] = is_na

        return new_x

    def transform(self, X):
        """
        For each column, it creates a new one indicating if that column is na
        """

        assert isinstance(X, pd.DataFrame) or isinstance(X, np.ndarray)

        if isinstance(X, pd.DataFrame):
            return self.transform_data_frame(X)

        if isinstance(X, np.ndarray):
            new_x = pd.DataFrame(X.copy())
            transformed_x = self.transform_data_frame(new_x)
            return transformed_x.values



# coding: utf-8


import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from category_encoders.ordinal import OrdinalEncoder
from category_encoders.m_estimate import MEstimateEncoder
import category_encoders.utils as util
from sklearn.utils.random import check_random_state

"""Nested target encoder"""

__author__ = "david26694 & cmougan"


class NestedTargetEncoder(BaseEstimator, util.TransformerWithTargetMixin):
    """Estimate of likelihood for nested data.

    This is a generalization of the m-probability estimate. The main difference
    is that instead of using a global prior, it can use a more fine-tuned prior.
    This only works for nested data. For instance, I have individuals who live
    in counties, that are inside states. If I want to estimate the likelihood
    encoding for a county, it is better to use as prior the estimate for the
    state instead of the global estimate.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop encoded columns with 0 variance.
    feature_mapping: dict
        dictionary representing the child - parent relationship. keys are children.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).
    handle_missing: str
        options are 'return_nan', 'error' and 'value', defaults to 'value', which returns the prior probability.
    handle_unknown: str
        options are 'return_nan', 'error' and 'value', defaults to 'value', which returns the prior probability.
    randomized: bool,
        adds normal (Gaussian) distribution noise into training data in order to decrease overfitting (testing data are untouched).
    sigma: float
        standard deviation (spread or "width") of the normal distribution.
    m_prior: float
        this is the "m" in the m-probability estimate for the global mean. Higher value of m results into stronger shrinking.
        It is used whenever we estimate a likelihood using the global mean as a prior.
        M is non-negative.
    m_parent: float
        this is the "m" in the m-probability estimate. Higher value of m results into stronger shrinking.
        It is used whenever we estimate a likelihood using the parent mean as a prior.
        M is non-negative.

    Example
    -------
    >>> from sktools import NestedTargetEncoder
    >>> import pandas as pd
    >>> X = pd.DataFrame(
    >>>     {
    >>>         "child": ["a", "a", "b", "b", "b", "c", "c", "d", "d", "d"],
    >>>         "parent": ["e", "e", "e", "e", "e", "f", "f", "f", "f", "f",]
    >>>     }
    >>> )
    >>> y = pd.Series([1, 2, 3, 1, 2, 4, 4, 5, 4, 4.5])
    >>> ne = NestedTargetEncoder(feature_mapping={"child": "parent"}, m_prior=0)
    >>> ne.fit_transform(X, y)
          child  parent
    0  2.016667     1.8
    1  2.016667     1.8
    2  2.262500     1.8
    3  2.262500     1.8
    4  2.262500     1.8
    5  3.683333     4.3
    6  3.683333     4.3
    7  4.137500     4.3
    8  4.137500     4.3
    9  4.137500     4.3

    References
    ----------

    .. [1] Additive smoothing, from https://en.wikipedia.org/wiki/Additive_smoothing#Generalized_to_the_case_of_known_incidence_rates

    """

    def __init__(
        self,
        verbose=0,
        cols=None,
        drop_invariant=False,
        feature_mapping={},
        return_df=True,
        handle_unknown="value",
        handle_missing="value",
        random_state=None,
        randomized=False,
        sigma=0.05,
        m_prior=1.0,
        m_parent=1.0,
    ):
        self.verbose = verbose
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.cols = cols
        self.feature_mapping = feature_mapping
        self.ordinal_encoder = None
        self._dim = None
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._sum = None
        self._count = None
        self.random_state = random_state
        self.randomized = randomized
        self.sigma = sigma
        self.m_prior = m_prior
        self.m_parent = m_parent
        self.feature_names = None
        self.parent_cols = None
        self.parent_encoder = None

    # noinspection PyUnusedLocal
    def fit(self, X, y, **kwargs):
        """Fit encoder according to X and binary y.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Binary target values.

        Returns
        -------

        self : encoder
            Returns self.

        """

        # Create parent encoder and fit it
        self.parent_cols = list(self.feature_mapping.values())
        self.parent_encoder = MEstimateEncoder(
            verbose=self.verbose,
            cols=self.parent_cols,
            drop_invariant=self.drop_invariant,
            return_df=self.return_df,
            handle_unknown=self.handle_unknown,
            handle_missing=self.handle_missing,
            random_state=self.random_state,
            randomized=self.randomized,
            sigma=self.sigma,
            m=self.m_prior,
        )
        self.parent_encoder.fit(X, y)

        # Unite parameters into pandas types
        X = util.convert_input(X)
        y = util.convert_input_vector(y, X.index).astype(float)

        # The lengths must be equal
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The length of X is "
                + str(X.shape[0])
                + " but length of y is "
                + str(y.shape[0])
                + "."
            )

        self._dim = X.shape[1]

        # If columns aren't passed, just use every string column
        if self.cols is None:
            self.cols = util.get_obj_cols(X)
        else:
            self.cols = util.convert_cols_to_list(self.cols)

        if self.handle_missing == "error":
            if X[self.cols].isnull().any().any():
                raise ValueError("Columns to be encoded can not contain null")

        # Check that children and parents are disjoint
        children = set(self.feature_mapping.keys())
        parents = set(self.feature_mapping.values())
        if len(children.intersection(parents)) > 0:
            raise ValueError("No column should be a child and a parent")

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose,
            cols=self.cols,
            handle_unknown="value",
            handle_missing="value",
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)

        # Training
        self.mapping = self._train(X_ordinal, y)

        X_temp = self.transform(X, override_return_df=True)
        self.feature_names = X_temp.columns.tolist()

        # Store column names with approximately constant variance on the training data
        if self.drop_invariant:
            self.drop_cols = []
            generated_cols = util.get_generated_cols(X, X_temp, self.cols)
            self.drop_cols = [
                x for x in generated_cols if X_temp[x].var() <= 10e-5
            ]
            try:
                [self.feature_names.remove(x) for x in self.drop_cols]
            except KeyError as e:
                if self.verbose > 0:
                    print(
                        "Could not remove column from feature names."
                        "Not found in generated cols.\n{}".format(e)
                    )
        return self

    def transform(self, X, y=None, override_return_df=False):
        """Perform the transformation to new categorical data.

        When the data are used for model training, it is important to also pass the target in order to apply leave one out.

        Parameters
        ----------

        X : array-like, shape = [n_samples, n_features]
        y : array-like, shape = [n_samples] when transform by leave one out
            None, when transform without target information (such as transform test set)


        Returns
        -------

        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """

        if self.handle_missing == "error":
            if X[self.cols].isnull().any().any():
                raise ValueError("Columns to be encoded can not contain null")

        if self._dim is None:
            raise ValueError(
                "Must train encoder before it can be used to transform data."
            )

        # Unite the input into pandas types
        X = util.convert_input(X)

        # Then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError(
                "Unexpected input dimension %d, expected %d"
                % (X.shape[1], self._dim,)
            )

        # If we are encoding the training data, we have to check the target
        if y is not None:
            y = util.convert_input_vector(y, X.index).astype(float)
            if X.shape[0] != y.shape[0]:
                raise ValueError(
                    "The length of X is "
                    + str(X.shape[0])
                    + " but length of y is "
                    + str(y.shape[0])
                    + "."
                )

        if not list(self.cols):
            return X

        # Do not modify the input argument
        X = X.copy(deep=True)

        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == "error":
            if X[self.cols].isin([-1]).any().any():
                raise ValueError("Unexpected categories found in dataframe")

        # Loop over the columns and replace the nominal values with the numbers
        X = self._score(X, y)

        # Postprocessing
        # Note: We should not even convert these columns.
        if self.drop_invariant:
            for col in self.drop_cols:
                X.drop(col, 1, inplace=True)

        if self.return_df or override_return_df:
            return X
        else:
            return X.values

    def _train(self, X, y):
        # Initialize the output
        mapping = {}

        # Calculate global statistics
        self._sum = y.sum()
        self._count = y.count()
        prior = self._sum / self._count

        for switch in self.ordinal_encoder.category_mapping:
            col = switch.get("col")
            values = switch.get("mapping")

            # Easy case, the child is not in the child - parent dictionary.
            # We just use the plain m-estimator with the global prior
            if col not in self.feature_mapping:
                stats = y.groupby(X[col]).agg(["sum", "count", "mean"])

                estimate = (stats["sum"] + prior * self.m_prior) / (
                    stats["count"] + self.m_prior
                )

            # Not so easy case, we have to deal with the parent
            else:
                parent_col = self.feature_mapping[col]

                # Check son-parent unique relation
                unique_parents = X.groupby([col]).agg({parent_col: "nunique"})[
                    parent_col
                ]

                more_1_parent = unique_parents[unique_parents > 1]

                if any(unique_parents > 1) and more_1_parent.index >= 0:
                    raise ValueError(
                        f"There are children with more than one parent, {more_1_parent}"
                    )

                # Get parent stats
                te_parent = self.parent_encoder.transform(X)[parent_col]
                parent_mapping = pd.DataFrame(
                    {"te_parent": te_parent, parent_col: X[parent_col]}
                ).drop_duplicates()

                # Compute child statistics
                stats = y.groupby(X[col]).agg(["sum", "count", "mean"])

                # Relate parent and child stats
                groups = X.loc[:, [parent_col, col]].drop_duplicates()

                stats = stats.merge(groups, how="left", on=col).merge(
                    parent_mapping, how="left", on=parent_col
                )

                # In case of numpy array
                stats = stats.rename(columns={"key_0": col})
                stats = stats.set_index(col)

                # Calculate the m-probability estimate using the parent prior
                estimate = (
                    stats["sum"] + stats["te_parent"] * self.m_parent
                ) / (stats["count"] + self.m_parent)

            # Ignore unique columns. This helps to prevent overfitting on id-like columns
            if len(stats["count"]) == self._count:
                estimate[:] = prior

            # Column doesn't have parent - handle imputation as always
            if col not in self.feature_mapping:
                if self.handle_unknown == "return_nan":
                    estimate.loc[-1] = np.nan
                elif self.handle_unknown == "value":
                    estimate.loc[-1] = prior

                if self.handle_missing == "return_nan":
                    estimate.loc[values.loc[np.nan]] = np.nan
                elif self.handle_missing == "value":
                    estimate.loc[-2] = prior
            # With parents - we leave the imputation for afterwards
            else:
                # Unknown
                estimate.loc[-1] = np.nan
                # Missing
                estimate.loc[values.loc[np.nan]] = np.nan

            # Store the m-probability estimate for transform() function
            mapping[col] = estimate

        return mapping

    def _score(self, X, y):

        X_parents = self.parent_encoder.transform(X)

        for col in self.cols:

            # Easy case - not having parents (as m estimator)
            if col not in self.feature_mapping:
                # Score the column
                X[col] = X[col].map(self.mapping[col])

            # Harder case - having parents
            else:

                # Split missing and unknown values
                unknown = X[col] == -1
                missing = X[col] == -2

                # Apply regular transformation
                X[col] = X[col].map(self.mapping[col].drop_duplicates())

                # Impute unknown with parent
                parent_col = self.feature_mapping[col]
                if self.handle_unknown == "value":
                    X[col] = X[col].mask(unknown, X_parents[parent_col])

                # Impute missing with parent
                if self.handle_missing == "value":
                    X[col] = X[col].mask(missing, X_parents[parent_col])

            # Randomization is meaningful only for training data -> we do it only if y is present
            if self.randomized and y is not None:
                random_state_generator = check_random_state(self.random_state)
                X[col] = X[col] * random_state_generator.normal(
                    1.0, self.sigma, X[col].shape[0]
                )

        return X

    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.

        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!

        """
        if not isinstance(self.feature_names, list):
            raise ValueError(
                "Estimator has to be fitted to return feature names."
            )
        else:
            return self.feature_names


"""Percentile Encoder"""

__author__ = "cmougan & david26694"


class PercentileEncoder(BaseEstimator, util.TransformerWithTargetMixin):
    """Percentile Encoding for categorical features.

    For the case of categorical target: features are replaced with a blend of
    posterior percentile of the target given particular categorical value and
    the prior probability of the target over all the training data.

    For the case of continuous target: features are replaced with a blend of the
    expected value of the target given particular categorical value and the
    expected value of the target over all the training data.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    return_df: bool
        boolean for whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).
    handle_missing: str
        options are 'error', 'return_nan'  and 'value', defaults to 'value', which returns the target percentile.
    handle_unknown: str
        options are 'error', 'return_nan' and 'value', defaults to 'value', which returns the target percentile.

    Example
    -------
    >>> from category_encoders import *
    >>> import pandas as pd
    >>> from sklearn.datasets import load_boston
    >>> bunch = load_boston()
    >>> y = bunch.target
    >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)
    >>> enc = PercentileEncoder(cols=['CHAS', 'RAD']).fit(X, y)
    >>> numeric_dataset = enc.transform(X)
    >>> print(numeric_dataset.info())
    <class 'pandas.core.frame.DataFrame'>
    RangeIndex: 506 entries, 0 to 505
    Data columns (total 13 columns):
    CRIM       506 non-null float64
    ZN         506 non-null float64
    INDUS      506 non-null float64
    CHAS       506 non-null float64
    NOX        506 non-null float64
    RM         506 non-null float64
    AGE        506 non-null float64
    DIS        506 non-null float64
    RAD        506 non-null float64
    TAX        506 non-null float64
    PTRATIO    506 non-null float64
    B          506 non-null float64
    LSTAT      506 non-null float64
    dtypes: float64(13)
    memory usage: 51.5 KB
    None

    References
    ----------

    """

    def __init__(
        self,
        verbose=0,
        cols=None,
        drop_invariant=False,
        return_df=True,
        handle_missing="value",
        handle_unknown="value",
        percentile=50,
    ):
        self.return_df = return_df
        self.drop_invariant = drop_invariant
        self.drop_cols = []
        self.verbose = verbose
        self.cols = cols
        self.ordinal_encoder = None
        self._dim = None
        self.mapping = None
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self._percentile = None
        self.feature_names = None
        self.percentile = percentile

    def fit(self, X, y, **kwargs):
        """Fit encoder according to X and y.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples
            and n_features is the number of features.
        y : array-like, shape = [n_samples]
            Target values.

        Returns
        -------
        self : encoder
            Returns self.

        """

        # unite the input into pandas types
        X = util.convert_input(X)
        y = util.convert_input_vector(y, X.index)

        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The length of X is "
                + str(X.shape[0])
                + " but length of y is "
                + str(y.shape[0])
                + "."
            )

        self._dim = X.shape[1]

        # if columns aren't passed, just use every string column
        if self.cols is None:
            self.cols = util.get_obj_cols(X)
        else:
            self.cols = util.convert_cols_to_list(self.cols)

        if self.handle_missing == "error":
            if X[self.cols].isnull().any().any():
                raise ValueError("Columns to be encoded can not contain null")

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose,
            cols=self.cols,
            handle_unknown="value",
            handle_missing="value",
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)
        X_ordinal = self.ordinal_encoder.transform(X)
        self.mapping = self.fit_percentile_encoding(X_ordinal, y)

        X_temp = self.transform(X, override_return_df=True)
        self.feature_names = list(X_temp.columns)

        if self.drop_invariant:
            self.drop_cols = []
            X_temp = self.transform(X)
            generated_cols = util.get_generated_cols(X, X_temp, self.cols)
            self.drop_cols = [
                x for x in generated_cols if X_temp[x].var() <= 10e-5
            ]
            try:
                [self.feature_names.remove(x) for x in self.drop_cols]
            except KeyError as e:
                if self.verbose > 0:
                    print(
                        "Could not remove column from feature names."
                        "Not found in generated cols.\n{}".format(e)
                    )

        return self

    def fit_percentile_encoding(self, X, y):
        mapping = {}

        for switch in self.ordinal_encoder.category_mapping:
            col = switch.get("col")
            values = switch.get("mapping")

            prior = self._percentile = np.percentile(y, self.percentile)

            stats = y.groupby(X[col]).apply(
                lambda x: np.percentile(x, self.percentile)
            )

            smoothing = stats

            if self.handle_unknown == "return_nan":
                smoothing.loc[-1] = np.nan
            elif self.handle_unknown == "value":
                smoothing.loc[-1] = prior

            if self.handle_missing == "return_nan":
                smoothing.loc[values.loc[np.nan]] = np.nan
            elif self.handle_missing == "value":
                smoothing.loc[-2] = prior

            mapping[col] = smoothing

        return mapping

    def transform(self, X, y=None, override_return_df=False):
        """Perform the transformation to new categorical data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
        y : array-like, shape = [n_samples] when transform by leave one out
            None, when transform without target info (such as transform test set)

        Returns
        -------
        p : array, shape = [n_samples, n_numeric + N]
            Transformed values with encoding applied.

        """

        if self.handle_missing == "error":
            if X[self.cols].isnull().any().any():
                raise ValueError("Columns to be encoded can not contain null")

        if self._dim is None:
            raise ValueError(
                "Must train encoder before it can be used to transform data."
            )

        # unite the input into pandas types
        X = util.convert_input(X)

        # then make sure that it is the right size
        if X.shape[1] != self._dim:
            raise ValueError(
                "Unexpected input dimension %d, expected %d"
                % (X.shape[1], self._dim,)
            )

        # if we are encoding the training data, we have to check the target
        if y is not None:
            y = util.convert_input_vector(y, X.index)
            if X.shape[0] != y.shape[0]:
                raise ValueError(
                    "The length of X is "
                    + str(X.shape[0])
                    + " but length of y is "
                    + str(y.shape[0])
                    + "."
                )

        if not list(self.cols):
            return X

        X = self.ordinal_encoder.transform(X)

        if self.handle_unknown == "error":
            if X[self.cols].isin([-1]).any().any():
                raise ValueError("Unexpected categories found in dataframe")

        X = self.percentile_encode(X)

        if self.drop_invariant:
            for col in self.drop_cols:
                X.drop(col, 1, inplace=True)

        if self.return_df or override_return_df:
            return X
        else:
            return X.values

    def percentile_encode(self, X_in):
        X = X_in.copy(deep=True)

        for col in self.cols:
            X[col] = X[col].map(self.mapping[col])

        return X

    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.

        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!

        """

        if not isinstance(self.feature_names, list):
            raise ValueError(
                "Must fit data first. Affected feature names are not known "
                "before."
            )
        else:
            return self.feature_names



"""Top-level package for sktools."""

__author__ = """David Masip Bonet"""
__email__ = "david26694@gmail.com"
__version__ = "0.1.0"

from .selectors import TypeSelector, ItemSelector
from .matrix_denser import MatrixDenser
from .imputer import IsEmptyExtractor
from .encoders import PercentileEncoder, NestedTargetEncoder
from .quantilegroups import (
    GroupedQuantileTransformer,
    PercentileGroupFeaturizer,
    MeanGroupFeaturizer,
)
